//////////////////////////////////////////////////////////////////////////
//
//  Copyright (c) 2008-2011, Image Engine Design Inc. All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without
//  modification, are permitted provided that the following conditions are
//  met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//
//     * Neither the name of Image Engine Design nor the names of any
//       other contributors to this software may be used to endorse or
//       promote products derived from this software without specific prior
//       written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
//  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
//  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
//  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
//  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
//  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
//  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
//  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//////////////////////////////////////////////////////////////////////////

#ifndef IECORE_CUBICBASIS_H
#define IECORE_CUBICBASIS_H

#include "IECore/Export.h"

IECORE_PUSH_DEFAULT_VISIBILITY
#include "OpenEXR/ImathMatrix.h"
IECORE_POP_DEFAULT_VISIBILITY

namespace IECore
{

enum class StandardCubicBasis
{
		Unknown,
		Linear,
		Bezier,
		BSpline,
		CatmullRom
};

/// Provides a basis matrix class for use in constructing cubic curves.
/// \ingroup mathGroup
template<typename T>
class IECORE_EXPORT CubicBasis
{

	public :

		typedef T BaseType;
		typedef Imath::Matrix44<T> MatrixType;

		MatrixType matrix;
		unsigned step;

		CubicBasis( const MatrixType &m, unsigned s );

		CubicBasis( StandardCubicBasis standardBasis );

		template<class S>
		inline void coefficients( S t, S &c0, S &c1, S &c2, S &c3 ) const;
		template<class S>
		inline void coefficients( S t, S c[4] ) const;

		inline int numCoefficients() const;

		template<class S>
		inline S operator() ( S t, S p0, S p1, S p2, S p3 ) const;
		template<class S>
		inline S operator() ( S t, const S p[4] ) const;
		template<class S>
		inline S operator() ( typename S::BaseType t, const S &p0, const S &p1, const S &p2, const S &p3 ) const;
		template<class S>
		inline S operator() ( typename S::BaseType t, const S p[4] ) const;

		//! @name Derivatives
		/// Methods for computing the first derivatives with respect to 't', which for a
		/// curve is also its tangent at 't'.
		////////////////////////////////////////////////////////////////////////////////////////
		//@{
		template<class S>
		inline void derivativeCoefficients( S t, S &c0, S &c1, S &c2, S &c3 ) const;
		template<class S>
		inline void derivativeCoefficients( S t, S c[4] ) const;

		template<class S>
		inline S derivative( S t, S p0, S p1, S p2, S p3 ) const;
		template<class S>
		inline S derivative( S t, const S p[4] ) const;
		template<class S>
		inline S derivative( typename S::BaseType t, const S &p0, const S &p1, const S &p2, const S &p3 ) const;
		template<class S>
		inline S derivative( typename S::BaseType t, const S p[4] ) const;
		//@}

		//! @name Integrals
		/// Methods for computing integrals
		////////////////////////////////////////////////////////////////////////////////////////
		//@{
		template<class S>
		inline void integralCoefficients( S t0, S t1, S &c0, S &c1, S &c2, S &c3 ) const;
		template<class S>
		inline void integralCoefficients( S t0, S t1, S c[4] ) const;

		template<class S>
		inline S integral( S t0, S t1, S p0, S p1, S p2, S p3 ) const;
		template<class S>
		inline S integral( S t0, S t1, const S p[4] ) const;
		template<class S>
		inline S integral( typename S::BaseType t0, typename S::BaseType t1, const S &p0, const S &p1, const S &p2, const S &p3 ) const;
		template<class S>
		inline S integral( typename S::BaseType t0, typename S::BaseType t1, const S p[4] ) const;
		//@}

		//! @name Critical points
		/// Methods for computing the critical points with respect to 't', which for a
		/// curve are points where it may change direction
		////////////////////////////////////////////////////////////////////////////////////////
		//@{
		template<class S>
		inline bool criticalPoints( const S p[4], S &p0, S &p1 ) const;
		//@}

		bool operator==( const CubicBasis &rhs ) const;
		bool operator!=( const CubicBasis &rhs ) const;

		static const CubicBasis &linear();
		static const CubicBasis &bezier();
		static const CubicBasis &bSpline();
		static const CubicBasis &catmullRom();

		StandardCubicBasis standardBasis() const;
};

typedef CubicBasis<float> CubicBasisf;
typedef CubicBasis<double> CubicBasisd;

} // namespace IECore

#include "IECore/CubicBasis.inl"

#endif // IECORE_CUBICBASIS_H
